using System;
using System.Collections;
using System.ComponentModel;

namespace BUGLists
{
	#region csCustomerList class
	/// <summary>
	/// Summary description for csCustomerList.
	/// </summary>
	public class csCustomerList: CollectionBase, IBindingList, IEnumerable
	{
		#region Filtering variables
		string _lastNameBeginsWith;
		#endregion

		#region Variables for test data generation
		string[]		_titleArray		= new string [ 3 ];
		string[]		_lastNameArray	= new string [ 5 ];
		System.Random	_random;
		#endregion
		
		#region Constructor
		public csCustomerList()
		{
			if ( _random == null )
			{
				_random = new System.Random();
			}
			LoadDataArrays();
			_lastNameBeginsWith = "";
		}
		#endregion

		#region Test data generation
		// Create a specified number of Customers & add them to the list
		public void GenerateTestData( int p_noOfCustomers )
		{
			for ( int l_customerNo = 0; l_customerNo < p_noOfCustomers; l_customerNo++ )
			{
				csCustomer l_newCustomer = new csCustomer( this );
				( ( IEditableObject )l_newCustomer ).BeginEdit();
				InitialiseCustomer( l_newCustomer, l_customerNo );
				( ( IList )this ).Add( l_newCustomer );
				( ( IEditableObject )l_newCustomer ).EndEdit();
			}
		}

		// Load some randomised data into a Customer
		public void InitialiseCustomer( ICustomer p_customer, int p_customerNo)
		{
			p_customer.Title = _titleArray[ _random.Next( 3 ) ];
			p_customer.FirstName = "Customer " + p_customerNo.ToString();
			p_customer.LastName = _lastNameArray[ _random.Next( 5 ) ];
		}

		// Load the arrays used for Customer initialisation
		private void LoadDataArrays()
		{
			_titleArray[ 0 ]	= "Mr";
			_titleArray[ 1 ]	= "Mrs";
			_titleArray[ 2 ]	= "Miss";
			_lastNameArray[ 0 ] = "Smith";
			_lastNameArray[ 1 ] = "Jones";
			_lastNameArray[ 2 ] = "Brown";
			_lastNameArray[ 3 ] = "Green";
			_lastNameArray[ 4 ] = "Cooper";
		}
		#endregion

		#region IBindingList Members

		public void AddIndex(PropertyDescriptor property)
		{
			// Do nothing
		}

		public bool AllowNew
		{
			get
			{
				return true;
			}
		}

		public void ApplySort(PropertyDescriptor property, System.ComponentModel.ListSortDirection direction)
		{
			throw new NotSupportedException();
		}

		public PropertyDescriptor SortProperty
		{
			get
			{
				throw new NotSupportedException();
			}
		}

		public int Find(PropertyDescriptor property, object key)
		{
			throw new NotSupportedException();
		}

		public bool SupportsSorting
		{
			get
			{
				return false;
			}
		}

		public bool IsSorted
		{
			get
			{
				return false;
			}
		}

		public bool AllowRemove
		{
			get
			{
				return true;
			}
		}

		public bool SupportsSearching
		{
			get
			{
				return false;
			}
		}

		public System.ComponentModel.ListSortDirection SortDirection
		{
			get
			{
				throw new NotSupportedException();
			}
		}

		public event System.ComponentModel.ListChangedEventHandler ListChanged;

		public bool SupportsChangeNotification
		{
			get
			{
				return true;
			}
		}

		public void RemoveSort()
		{
			throw new NotSupportedException();
		}

		public object AddNew()
		{
			csCustomer l_newCustomer = new csCustomer( this );
			( ( IList )this ).Add( l_newCustomer );
			return l_newCustomer;
		}

		public bool AllowEdit
		{
			get
			{
				return true;
			}
		}

		public void RemoveIndex(PropertyDescriptor property)
		{
			// Do nothing
		}

		// Helper function for event handling
		private void OnListChanged( object p_sender, ListChangedEventArgs p_eventArgs )
		{
			if ( ListChanged != null )
			{
				ListChanged( p_sender, p_eventArgs );
			}
		}
		#endregion

		#region Strongly typed list management methods
		public int Add( ICustomer p_newCustomer )
		{
			return List.Add( p_newCustomer );
		}

		public bool Contains( ICustomer p_customer )
		{
			return List.Contains( p_customer );
		}

		public int IndexOf( ICustomer p_customer )
		{
			return List.IndexOf( p_customer );
		}

		public void Insert( int p_position, ICustomer p_customer )
		{
			List.Insert( p_position, p_customer );
		}

		public void Remove( ICustomer p_customer )
		{
			List.Remove( p_customer );
		}
		#endregion

		#region CollectionBase event overrides
		protected override void OnValidate(object value)
		{
			if ( ! ( value is ICustomer ) )
			{
				throw new ArgumentException( "You may only put an ICustomer in a csCustomerList" );
			}
		}

		protected override void OnClearComplete()
		{
			OnListChanged( this, new ListChangedEventArgs( ListChangedType.Reset, 0 ) );
		}

		protected override void OnInsertComplete(int index, object value)
		{
			OnListChanged( this, new ListChangedEventArgs( ListChangedType.ItemAdded, index ) );
		}

		protected override void OnRemoveComplete(int index, object value)
		{
			OnListChanged( this, new ListChangedEventArgs( ListChangedType.ItemDeleted, index ) );
		}

		protected override void OnSetComplete(int index, object oldValue, object newValue)
		{
			OnListChanged( this, new ListChangedEventArgs( ListChangedType.ItemChanged, index ) );
		}
		#endregion
		
		#region Filtering stuff
		public string LastNameBeginsWith
		{
			get
			{
				return _lastNameBeginsWith;
			}
			set
			{
				_lastNameBeginsWith = value;
			}
		}
		#endregion

		#region IEnumerable Members

		public new IEnumerator GetEnumerator()
		{
			return new csCustomerListEnumerator( this, LastNameBeginsWith );
		}

		#endregion
		
		#region Sorting
		public void Sort()
		{
			this.InnerList.Sort();
			OnListChanged( this, new ListChangedEventArgs( ListChangedType.Reset, 0 ) );
		}

		public void Reverse()
		{
			this.InnerList.Reverse();
			OnListChanged( this, new ListChangedEventArgs( ListChangedType.Reset, 0 ) );
		}

		public void Sort( IComparer p_comparer )
		{
			this.InnerList.Sort( p_comparer );
			OnListChanged( this, new ListChangedEventArgs( ListChangedType.Reset, 0 ) );
		}
		#endregion
	}
	#endregion

	#region csCustomerListEnumerator class
	public class csCustomerListEnumerator: IEnumerator
	{
		#region Enumeration variables
		int _position;
		csCustomerList _customers;
		#endregion

		#region Filtering variables
		string _lastNameBeginsWith;
		#endregion
		
		#region Constructor
		public csCustomerListEnumerator( csCustomerList p_customers, string p_lastNameBeginsWith )
		{
			_customers = p_customers;
			_position = -1;
			_lastNameBeginsWith = p_lastNameBeginsWith;
		}
		#endregion

		#region Helper function
		private csCustomer CurrentCustomer()
		{
			return ( ( IList )_customers )[ _position ] as csCustomer;
		}
		#endregion

		#region IEnumerator members
		object IEnumerator.Current
		{
			get
			{
				return CurrentCustomer();
			}
		}

		bool IEnumerator.MoveNext()
		{
			// Test to see if there is another element of the list after the current one
			if ( _position < ( _customers.Count - 1 ) )
			{
				int l_oldPosition = _position;
				_position = _position + 1;
				// Do filtering, if required
				if ( _lastNameBeginsWith.Length > 0 )
				{
					// Check the start of the Last Name against the comparison string (case insensitive)
					// If they don't match, move to the next Customer, until we reach the end of the list
					while ( ( _position < _customers.Count ) && 
						( string.Compare( ( ( ICustomer )CurrentCustomer() ).LastName, 0, _lastNameBeginsWith, 0, _lastNameBeginsWith.Length, true ) != 0 ) )
					{
						_position = _position + 1;
					}
				}
				bool l_result = _position < _customers.Count;
				if ( ! l_result )
				{
					// If there isn't another Customer that meets the filtering criteria,
					// restore the old position
					_position = l_oldPosition;
				}
				return l_result;
			}
			else
			{
				return false;
			}
		}

		void IEnumerator.Reset()
		{
			_position = -1;
		}
		#endregion
	}
	#endregion
}
